# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Distances computation and comparison of kapture objects
"""

import math
import numpy as np
from typing import Dict, List, Optional, Set, Tuple, Any, Union
import inspect

import kapture
from kapture import flatten
from kapture.core.Sensors import Camera
from kapture.utils.logging import getLogger

from .pose_operations import pose_transform_distance


def float_iszero(distance: float, threshold: float = 1e-05) -> bool:
    """
    Computes if a distance is close to zero modulo an epsilon.

    :param distance: distance to evaluate
    :param threshold: the epsilon value
    :return: true or false if the condition is met
    """
    return math.isclose(distance, 0.0, rel_tol=threshold, abs_tol=threshold)


def is_distance_within_threshold(
        pose_distance: Tuple[float, float],
        pose_thresholds: Tuple[float, float] = (1e-05, 1e-05)
) -> bool:
    """
    compare a pose distance tuple to (0,0) with some thresholds

    :param pose_distance: (translation_distance, rotation_distance)
    :type pose_distance: Tuple[float, float]
    :param pose_thresholds: (translation_threshold, rotation_threshold)
    :type pose_thresholds: Tuple[float, float], optional
    :return: True if both rotation and translation distance are within threshold
    :rtype: bool
    """
    translation_distance, rotation_distance = pose_distance
    translation_threshold, rotation_threshold = pose_thresholds
    return float_iszero(translation_distance, translation_threshold)\
        and float_iszero(rotation_distance, rotation_threshold)


def equal_poses(pose_a: kapture.PoseTransform, pose_b: kapture.PoseTransform) -> bool:
    """
    Compare the two pose to check if they are equal.

    :param pose_a: first pose
    :param pose_b: second pose
    :return: True if they are equal, False otherwise
    """
    pose_a_nones = (pose_a.r is None, pose_a.t is None)
    pose_b_nones = (pose_b.r is None, pose_b.t is None)
    if pose_a_nones != pose_b_nones:
        return False

    pose_distance = pose_transform_distance(pose_a, pose_b)
    if pose_a_nones == (True, True):  # a and b have None rotation, None translation
        return True
    elif pose_a_nones == (True, False):  # a and b have None rotation, valid translation
        return float_iszero(pose_distance[0])
    elif pose_a_nones == (False, True):  # a and b have valid rotation, None translation
        return float_iszero(pose_distance[1])
    else:
        return is_distance_within_threshold(pose_distance)


def equal_camera_params(camera_params_a: List[float], camera_params_b: List[float]) -> bool:
    """
    Checks if the camera parameters are equals.

    :param camera_params_a: first camera parameters
    :param camera_params_b: second camera parameters
    :return: True if they are equal, False otherwise
    """
    return np.isclose([float(v) for v in camera_params_a], [float(v) for v in camera_params_b]).all()


def equal_sensors(sensors_a: Optional[kapture.Sensors], sensors_b: Optional[kapture.Sensors]) -> bool:
    """
    Compare two instances of kapture.Sensors.
    model_params for cameras are considered equal if np.isclose says so.

    :param sensors_a: first sensor definition
    :param sensors_b: second sensor definition
    :return: True if they are identical, False otherwise.
    """
    if sensors_a is None and sensors_b is None:
        return True
    elif sensors_a is None and sensors_b is not None:
        return False
    elif sensors_a is not None and sensors_b is None:
        return False

    flattened_a = list(flatten(sensors_a, is_sorted=True))
    flattened_b = list(flatten(sensors_b, is_sorted=True))
    if len(flattened_a) != len(flattened_b):
        getLogger().debug('equal_sensors: a and b do not have the same number of elements')
        return False
    for (sensor_id_a, sensor_a), (sensor_id_b, sensor_b) in zip(flattened_a, flattened_b):
        # handling special case: name_a='' and name_b=None
        equal_id = sensor_id_a == sensor_id_b
        equal_name = (not sensor_a.name and not sensor_b.name) or (sensor_a.name == sensor_b.name)
        equal_type = sensor_a.sensor_type == sensor_b.sensor_type

        if not equal_id or not equal_name or not equal_type:
            getLogger().debug(
                f'equal_sensors: ({sensor_id_a}, {sensor_a}) != ({sensor_id_b}, {sensor_b})')
            return False

        equal_params = False
        if sensor_a.sensor_type in kapture.ALL_CAMERA_SENSOR_TYPES:
            assert isinstance(sensor_a, Camera)
            assert isinstance(sensor_b, Camera)
            if sensor_a.sensor_type != sensor_b.sensor_type:
                return False
            if sensor_a.camera_type == sensor_b.camera_type:
                equal_params = equal_camera_params(sensor_a.camera_params, sensor_b.camera_params)
        else:
            equal_params = sensor_a.sensor_params == sensor_b.sensor_params

        if not equal_params:
            getLogger().debug(
                f'equal_sensors: ({sensor_id_a}, {sensor_a}) != ({sensor_id_b}, {sensor_b})')
            return False
    return True


def equal_rigs(rigs_a: Optional[kapture.Rigs], rigs_b: Optional[kapture.Rigs]) -> bool:
    """
    Compare two instances of kapture.Rigs.
    Poses are compared with is_distance_within_threshold(pose_transform_distance())

    :param rigs_a: first set of rigs
    :param rigs_b: second set of rigs
    :return: True if they are identical, False otherwise.
    """
    if rigs_a is None and rigs_b is None:
        return True
    elif rigs_a is None and rigs_b is not None:
        return False
    elif rigs_a is not None and rigs_b is None:
        return False

    flattened_a = list(flatten(rigs_a, is_sorted=True))
    flattened_b = list(flatten(rigs_b, is_sorted=True))
    if len(flattened_a) != len(flattened_b):
        getLogger().debug('equal_rigs: a and b do not have the same number of elements')
        return False
    for (rig_id_a, sensor_id_a, pose_a), (rig_id_b, sensor_id_b, pose_b) in zip(flattened_a, flattened_b):
        if rig_id_a != rig_id_b or sensor_id_a != sensor_id_b:
            getLogger().debug(
                f'equal_rigs: ({rig_id_a}, {sensor_id_a}, {pose_a.r_raw}, {pose_a.t_raw}) !='
                f' ({rig_id_b}, {sensor_id_b}, {pose_b.r_raw}, {pose_b.t_raw})')
            return False
        if not equal_poses(pose_a, pose_b):
            getLogger().debug(
                f'equal_rigs: ({rig_id_a}, {sensor_id_a}, {pose_a.r_raw}, {pose_a.t_raw}) '
                f'is not close to '
                f'({rig_id_b}, {sensor_id_b}, {pose_b.r_raw}, {pose_b.t_raw})')
            return False
    return True


def equal_trajectories(
        trajectories_a: Optional[kapture.Trajectories],
        trajectories_b: Optional[kapture.Trajectories]) -> bool:
    """
    Compare two instances of kapture.Trajectories.
    Poses are compared with is_distance_within_threshold(pose_transform_distance())

    :param trajectories_a: first trajectory
    :param trajectories_b: second trajectory
    :return: True if they are identical, False otherwise.
    """
    if trajectories_a is None and trajectories_b is None:
        return True
    elif trajectories_a is None and trajectories_b is not None:
        return False
    elif trajectories_a is not None and trajectories_b is None:
        return False

    flattened_a = list(flatten(trajectories_a, is_sorted=True))
    flattened_b = list(flatten(trajectories_b, is_sorted=True))
    if len(flattened_a) != len(flattened_b):
        getLogger().debug('equal_trajectories: a and b do not have the same number of elements')
        return False
    for (timestamp_a, sensor_id_a, pose_a), (timestamp_b, sensor_id_b, pose_b) in zip(flattened_a, flattened_b):
        if timestamp_a != timestamp_b or sensor_id_a != sensor_id_b:
            getLogger().debug(
                f'equal_trajectories: ({timestamp_a}, {sensor_id_a}, {pose_a.r_raw}, {pose_a.t_raw}) !='
                f' ({timestamp_b}, {sensor_id_b}, {pose_b.r_raw}, {pose_b.t_raw})')
            return False
        if not equal_poses(pose_a, pose_b):
            getLogger().debug(
                f'equal_trajectories: ({timestamp_a}, {sensor_id_a}, {pose_a.r_raw}, {pose_a.t_raw}) '
                f'is not close to '
                f'({timestamp_b}, {sensor_id_b}, {pose_b.r_raw}, {pose_b.t_raw})')
            return False
    return True


def log_difference(a: List[Tuple[Any, ...]], b: List[Tuple[Any, ...]], func_name: str, trim_count: int = 5) -> None:
    """
    Records in the logger the difference between two values.

    :param a: first value
    :param b: second value
    :param func_name: comparison function to print
    :param trim_count: maximum number of values to record
    """
    if len(a) != len(b):
        getLogger().debug(f'{func_name}: a and b do not have the same number of elements')
    else:
        diffs = [(va, vb) for va, vb in zip(a, b) if va != vb]
        diffs = diffs[:trim_count]
        diffs = ['({}) != ({})'.format(', '.join([str(f) for f in va]),
                                       ', '.join([str(f) for f in vb]))
                 for va, vb in diffs]
        getLogger().debug('{}:\n{}'.format(func_name, '\n'.join(diffs)))


def equal_nested_dict_or_set(data_a, data_b, name_to_log, expected_type=None) -> bool:
    """
    Compare two instances of dictionary or set

    :return: True if they are identical, False otherwise.
    """
    if expected_type is not None:  # do type checking
        for data_x in [data_a, data_b]:
            if data_x is not None and not isinstance(data_x, expected_type):
                raise TypeError(f'expecting type {expected_type} in {name_to_log} (got {type(data_x)})')

    # early check if one (or both) are None
    if data_a is None and data_b is None:
        return True
    elif data_a is None and data_b is not None:
        return False
    elif data_a is not None and data_b is None:
        return False

    # check values
    flattened_a = list(flatten(data_a, is_sorted=True))
    flattened_b = list(flatten(data_b, is_sorted=True))
    are_equal = (flattened_a == flattened_b)
    if not are_equal:
        log_difference(flattened_a, flattened_b, name_to_log)
    return are_equal


def equal_sets(data_a: Set[Any], data_b: Set[Any], func_name: str = 'equal_sets'):
    set_difference = data_a.difference(data_b)
    are_equal = len(set_difference) == 0
    if not are_equal:
        getLogger().debug('{}:\n{}'.format(func_name, '\n'.join((str(s) for s in set_difference))))
    return are_equal


def equal_keypoints(
        data_a: kapture.Keypoints,
        data_b: kapture.Keypoints
) -> bool:
    """
    Compare two instances of kapture keypoints.

    :param data_a: first set of features
    :param data_b: second set of features
    :return: True if they are identical, False otherwise.
    """
    assert isinstance(data_a, kapture.Keypoints)
    assert isinstance(data_b, kapture.Keypoints)
    if data_a.type_name != data_b.type_name or data_a.dtype != data_b.dtype or data_a.dsize != data_b.dsize:
        return False
    current_function_name = inspect.getframeinfo(inspect.currentframe()).function
    return equal_sets(data_a, data_b, current_function_name)


def equal_keypoints_collections(data_a: Optional[Dict[str, kapture.Keypoints]],
                                data_b: Optional[Dict[str, kapture.Keypoints]]):
    """
    Compare two keypoints collections.
    :param data_a: first collection of features
    :param data_b: second collection of features
    :return: True if they are identical, False otherwise.
    """
    # early check if one is None
    if data_a is None and data_b is None:
        return True
    elif data_a is None and data_b is not None:
        return False
    elif data_a is not None and data_b is None:
        return False

    # should not happen because of previous lines, use assert to help your ide figure out the type of data
    assert data_a is not None
    assert data_b is not None

    current_function_name = inspect.getframeinfo(inspect.currentframe()).function
    if not equal_sets(set(data_a.keys()), set(data_b.keys()), current_function_name):
        return False
    for keypoint_type in data_a.keys():
        if not equal_keypoints(data_a[keypoint_type], data_b[keypoint_type]):
            return False
    return True


def equal_descriptors(
        data_a: kapture.Descriptors,
        data_b: kapture.Descriptors
) -> bool:
    """
    Compare two instances of kapture descriptors.

    :param data_a: first set of features
    :param data_b: second set of features
    :return: True if they are identical, False otherwise.
    """
    assert isinstance(data_a, kapture.Descriptors)
    assert isinstance(data_b, kapture.Descriptors)
    if data_a.type_name != data_b.type_name or data_a.dtype != data_b.dtype or data_a.dsize != data_b.dsize \
            or data_a.keypoints_type != data_b.keypoints_type or data_a.metric_type != data_b.metric_type:
        return False
    current_function_name = inspect.getframeinfo(inspect.currentframe()).function
    return equal_sets(data_a, data_b, current_function_name)


def equal_descriptors_collections(data_a: Optional[Dict[str, kapture.Descriptors]],
                                  data_b: Optional[Dict[str, kapture.Descriptors]]):
    """
    Compare two descriptors collections.
    :param data_a: first collection of features
    :param data_b: second collection of features
    :return: True if they are identical, False otherwise.
    """
    # early check if one is None
    if data_a is None and data_b is None:
        return True
    elif data_a is None and data_b is not None:
        return False
    elif data_a is not None and data_b is None:
        return False

    # should not happen because of previous lines, use assert to help your ide figure out the type of data
    assert data_a is not None
    assert data_b is not None

    current_function_name = inspect.getframeinfo(inspect.currentframe()).function
    if not equal_sets(set(data_a.keys()), set(data_b.keys()), current_function_name):
        return False
    for descriptors_type in data_a.keys():
        if not equal_descriptors(data_a[descriptors_type], data_b[descriptors_type]):
            return False
    return True


def equal_global_features(
        data_a: kapture.GlobalFeatures,
        data_b: kapture.GlobalFeatures
) -> bool:
    """
    Compare two instances of kapture global features.

    :param data_a: first set of features
    :param data_b: second set of features
    :return: True if they are identical, False otherwise.
    """
    assert isinstance(data_a, kapture.GlobalFeatures)
    assert isinstance(data_b, kapture.GlobalFeatures)
    if data_a.type_name != data_b.type_name or data_a.dtype != data_b.dtype or data_a.dsize != data_b.dsize \
            or data_a.metric_type != data_b.metric_type:
        return False
    current_function_name = inspect.getframeinfo(inspect.currentframe()).function
    return equal_sets(data_a, data_b, current_function_name)


def equal_global_features_collections(data_a: Optional[Dict[str, kapture.GlobalFeatures]],
                                      data_b: Optional[Dict[str, kapture.GlobalFeatures]]):
    """
    Compare two global_features collections.
    :param data_a: first collection of features
    :param data_b: second collection of features
    :return: True if they are identical, False otherwise.
    """
    # early check if one is None
    if data_a is None and data_b is None:
        return True
    elif data_a is None and data_b is not None:
        return False
    elif data_a is not None and data_b is None:
        return False

    # should not happen because of previous lines, use assert to help your ide figure out the type of data
    assert data_a is not None
    assert data_b is not None

    current_function_name = inspect.getframeinfo(inspect.currentframe()).function
    if not equal_sets(set(data_a.keys()), set(data_b.keys()), current_function_name):
        return False
    for global_features_type in data_a.keys():
        if not equal_global_features(data_a[global_features_type], data_b[global_features_type]):
            return False
    return True


def equal_records_camera(
        records_a: Optional[kapture.RecordsCamera],
        records_b: Optional[kapture.RecordsCamera]) -> bool:
    """
    Compare two instances of kapture.RecordsCamera.

    :param records_a: first set of records
    :param records_b: second set of records
    :return: True if they are identical, False otherwise.
    """
    expected_type = kapture.RecordsCamera
    current_function_name = inspect.getframeinfo(inspect.currentframe()).function
    return equal_nested_dict_or_set(records_a, records_b, current_function_name, expected_type)


def equal_records_lidar(
        records_a: Optional[kapture.RecordsLidar],
        records_b: Optional[kapture.RecordsLidar]) -> bool:
    """
    Compare two instances of kapture.RecordsLidar.

    :param records_a: first set of records
    :param records_b: second set of records
    :return: True if they are identical, False otherwise.
    """
    expected_type = kapture.RecordsLidar
    current_function_name = inspect.getframeinfo(inspect.currentframe()).function
    return equal_nested_dict_or_set(records_a, records_b, current_function_name, expected_type)


def equal_records_wifi(
        records_a: Optional[kapture.RecordsWifi],
        records_b: Optional[kapture.RecordsWifi]) -> bool:
    """
    Compare two instances of kapture.RecordsWifi.

    :param records_a: first set of records
    :param records_b: second set of records
    :return: True if they are identical, False otherwise.
    """
    expected_type = kapture.RecordsWifi
    current_function_name = inspect.getframeinfo(inspect.currentframe()).function
    return equal_nested_dict_or_set(records_a, records_b, current_function_name, expected_type)


def equal_records_bluetooth(
        records_a: Optional[kapture.RecordsBluetooth],
        records_b: Optional[kapture.RecordsBluetooth]) -> bool:
    """
    Compare two instances of kapture.RecordsBluetooth.

    :param records_a: first set of records
    :param records_b: second set of records
    :return: True if they are identical, False otherwise.
    """
    expected_type = kapture.RecordsBluetooth
    current_function_name = inspect.getframeinfo(inspect.currentframe()).function
    return equal_nested_dict_or_set(records_a, records_b, current_function_name, expected_type)


def equal_records_gnss(
        records_a: Optional[kapture.RecordsGnss],
        records_b: Optional[kapture.RecordsGnss]) -> bool:
    """
    Compare two instances of kapture.RecordsGnss.

    :param records_a: first set of records
    :param records_b: second set of records
    :return: True if they are identical, False otherwise.
    """
    expected_type = kapture.RecordsGnss
    current_function_name = inspect.getframeinfo(inspect.currentframe()).function
    return equal_nested_dict_or_set(records_a, records_b, current_function_name, expected_type)


def equal_records_accelerometer(
        records_a: Optional[kapture.RecordsAccelerometer],
        records_b: Optional[kapture.RecordsAccelerometer]) -> bool:
    """
    Compare two instances of kapture.RecordsAccelerometer.

    :param records_a: first set of records
    :param records_b: second set of records
    :return: True if they are identical, False otherwise.
    """
    expected_type = kapture.RecordsAccelerometer
    current_function_name = inspect.getframeinfo(inspect.currentframe()).function
    return equal_nested_dict_or_set(records_a, records_b, current_function_name, expected_type)


def equal_records_gyroscope(
        records_a: Optional[kapture.RecordsGyroscope],
        records_b: Optional[kapture.RecordsGyroscope]) -> bool:
    """
    Compare two instances of kapture.RecordsGyroscope.

    :param records_a: first set of records
    :param records_b: second set of records
    :return: True if they are identical, False otherwise.
    """
    expected_type = kapture.RecordsGyroscope
    current_function_name = inspect.getframeinfo(inspect.currentframe()).function
    return equal_nested_dict_or_set(records_a, records_b, current_function_name, expected_type)


def equal_records_magnetic(
        records_a: Optional[kapture.RecordsMagnetic],
        records_b: Optional[kapture.RecordsMagnetic]) -> bool:
    """
    Compare two instances of kapture.RecordsMagnetic.

    :param records_a: first set of records
    :param records_b: second set of records
    :return: True if they are identical, False otherwise.
    """
    expected_type = kapture.RecordsMagnetic
    current_function_name = inspect.getframeinfo(inspect.currentframe()).function
    return equal_nested_dict_or_set(records_a, records_b, current_function_name, expected_type)


def equal_matches(
        matches_a: kapture.Matches,
        matches_b: kapture.Matches) -> bool:
    """
    Compare two instances of kapture.Matches.

    :param matches_a: first set of matches
    :param matches_b: second set of matches
    :return: True if they are identical, False otherwise.
    """
    assert isinstance(matches_a, kapture.Matches)
    assert isinstance(matches_b, kapture.Matches)
    current_function_name = inspect.getframeinfo(inspect.currentframe()).function
    return equal_sets(matches_a, matches_b, current_function_name)


def equal_matches_collections(data_a: Optional[Dict[str, kapture.Matches]],
                              data_b: Optional[Dict[str, kapture.Matches]]):
    """
    Compare two matches collections.
    :param data_a: first collection of matches
    :param data_b: second collection of matches
    :return: True if they are identical, False otherwise.
    """
    # early check if one is None
    if data_a is None and data_b is None:
        return True
    elif data_a is None and data_b is not None:
        return False
    elif data_a is not None and data_b is None:
        return False

    # should not happen because of previous lines, use assert to help your ide figure out the type of data
    assert data_a is not None
    assert data_b is not None

    current_function_name = inspect.getframeinfo(inspect.currentframe()).function
    if not equal_sets(set(data_a.keys()), set(data_b.keys()), current_function_name):
        return False
    for keypoints_type in data_a.keys():
        if not equal_matches(data_a[keypoints_type], data_b[keypoints_type]):
            return False
    return True


def equal_observations(
        data_a: Optional[kapture.Observations],
        data_b: Optional[kapture.Observations]) -> bool:
    """
    Compare two instances of kapture.Observations.

    :param data_a: first set of observations
    :param data_b: second set of observations
    :return: True if they are identical, False otherwise.
    """
    expected_type = kapture.Observations
    current_function_name = inspect.getframeinfo(inspect.currentframe()).function
    return equal_nested_dict_or_set(data_a, data_b, current_function_name, expected_type)


def equal_points3d(
        points3d_a: Optional[kapture.Points3d],
        points3d_b: Optional[kapture.Points3d]) -> bool:
    """
    Compare two instances of kapture.Points3d.

    :param points3d_a: first set of points3d
    :param points3d_b: second set of points3d
    :return: True if they are identical, False otherwise.
    """
    if points3d_a is None and points3d_b is None:
        return True
    elif points3d_a is None and points3d_b is not None:
        return False
    elif points3d_a is not None and points3d_b is None:
        return False

    # ide guidance
    assert points3d_a is not None
    assert points3d_b is not None

    if len(points3d_a) != len(points3d_b):
        getLogger().debug('equal_points3d: a and b do not have the same number of elements')

    # internally converted to array of bool which cannot be a point3d
    bool_array = np.isclose(points3d_a.as_array(), points3d_b.as_array())
    are_equal = bool_array.all()
    if not are_equal:
        diffs = [n for n, b in enumerate(bool_array) if not b]
        diffs = diffs[:15]
        diffs = ['element {} : {} != {}'.format(n, points3d_a[n], points3d_b[n]) for n in diffs]
        getLogger().debug('equal_points3d:\n{}'.format('\n'.join(diffs)))
    return are_equal


def equal_kapture(
        data_a: kapture.Kapture,
        data_b: kapture.Kapture) -> bool:  # noqa: C901
    """
    Compare two instances of Kapture.
     Poses are compared with is_distance_within_threshold(pose_transform_distance())

    :param data_a: first kapture
    :param data_b: second kapture
    :return: True if they are identical, False otherwise.
    """
    # compare sensors
    if not equal_sensors(data_a.sensors, data_b.sensors):
        return False

    # compare rigs
    if not equal_rigs(data_a.rigs, data_b.rigs):
        return False

    # compare trajectories
    if not equal_trajectories(data_a.trajectories, data_b.trajectories):
        return False

    # compare records
    for record_sensor in ['camera', 'lidar', 'wifi', 'bluetooth', 'gnss', 'accelerometer', 'gyroscope', 'magnetic']:
        record_name = f'records_{record_sensor}'
        equal_record_function = eval(f'equal_records_{record_sensor}')
        if not equal_record_function(getattr(data_a, record_name), getattr(data_b, record_name)):
            return False

    # compare image features (keypoints, descriptors, global_features)
    if not equal_keypoints_collections(data_a.keypoints, data_b.keypoints):
        return False
    if not equal_descriptors_collections(data_a.descriptors, data_b.descriptors):
        return False
    if not equal_global_features_collections(data_a.global_features, data_b.global_features):
        return False

    # compare matches
    if not equal_matches_collections(data_a.matches, data_b.matches):
        return False

    # compare observations
    if not equal_observations(data_a.observations, data_b.observations):
        return False

    # compare points3d
    if not equal_points3d(data_a.points3d, data_b.points3d):
        return False
    return True



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Inject GPS to EXIF
"""
from typing import Dict, Optional
import logging
from PIL import Image
import piexif
from fractions import Fraction
import kapture
from kapture.io.records import get_image_fullpath

logger = logging.getLogger('exif')


def write_exif(image_filepath: str,
               exif_data):
    """
    read_data reverse operation: convert back string to ID, and write it to file.

    """
    assert exif_data is not None
    image = Image.open(image_filepath)
    exif_bytes = piexif.dump(exif_data)
    image.save(image_filepath, exif=exif_bytes)


def update_exif(image_filepath: str,
                exif_data):
    exif_bytes = piexif.dump(exif_data)
    piexif.insert(exif_bytes, image_filepath)


def clear_exif(image_filepath: str):
    piexif.remove(image_filepath)


def convert_dms_from_float(value, loc):
    """
    convert floats coordinates into degrees, munutes and seconds + ref tuple

    return: tuple like (25, 13, 48.343 ,'N')
    """
    if value < 0:
        loc_value = loc[0]
    elif value > 0:
        loc_value = loc[1]
    else:
        loc_value = ""
    abs_value = abs(value)
    deg = int(abs_value)
    t1 = (abs_value-deg) * 60
    min = int(t1)
    sec = round((t1 - min) * 60, 5)
    return deg, min, sec, loc_value


def convert_rational_from_float(number):
    """
    converts a float to rational as form of a tuple.
    """
    f = Fraction(str(number))  # str act as a round
    return f.numerator, f.denominator


def gps_record_to_exif_dict(
        gps_record: kapture.RecordGnss
) -> dict:
    """
    gps_record = x, y, z, utc, dop
    """

    lat_deg = convert_dms_from_float(gps_record.y, ["S", "N"])
    lng_deg = convert_dms_from_float(gps_record.x, ["W", "E"])
    latitude_hms = (convert_rational_from_float(lat_deg[0]), convert_rational_from_float(lat_deg[1]),
                    convert_rational_from_float(lat_deg[2]))
    longitude_hms = (convert_rational_from_float(lng_deg[0]), convert_rational_from_float(lng_deg[1]),
                     convert_rational_from_float(lng_deg[2]))
    gps_infos = {
            piexif.GPSIFD.GPSVersionID: (2, 0, 0, 0),
            piexif.GPSIFD.GPSLongitude: longitude_hms,
            piexif.GPSIFD.GPSLongitudeRef: b'N',
            piexif.GPSIFD.GPSLatitude: latitude_hms,
            piexif.GPSIFD.GPSLatitudeRef: b'E',
            piexif.GPSIFD.GPSAltitude: convert_rational_from_float(round(gps_record.z, 1)),
            piexif.GPSIFD.GPSAltitudeRef: 1,
            piexif.GPSIFD.GPSTimeStamp: convert_rational_from_float(gps_record.utc),
            piexif.GPSIFD.GPSDOP: convert_rational_from_float(gps_record.dop)
    }
    exif_data = {'GPS': gps_infos}
    return exif_data


def export_gps_to_exif(
        kapture_data: kapture.Kapture,
        kapture_dirpath: str,
        gps_id_to_cam_id: Optional[Dict[str, str]] = None
):
    """
    Export GPS from GNSS data in kapture, to image exifs.
    If the mapping from gnns_id to camera_id is not given,
    gnss_id is assumed to be to be gps_<cam_id>.
    """

    # sanity check
    if None in [kapture_data.sensors, kapture_data.records_camera, kapture_data.records_gnss]:
        logger.warning('cannot export exif: missing data.')
        return

    # auto build GPS/camera map, based on the prefix rule.
    if gps_id_to_cam_id is None:
        camera_ids = {cam_id
                      for cam_id, sensor in kapture_data.sensors.items()
                      if sensor.sensor_type == kapture.SENSOR_TYPE_CAMERA}

        gps_ids = {gps_id
                   for gps_id, sensor in kapture_data.sensors.items()
                   if sensor.sensor_type == 'gnss' and gps_id.startswith('GPS_')}

        gps_id_to_cam_id = {'GPS_' + cam_id: cam_id
                            for cam_id in camera_ids
                            if 'GPS_' + cam_id in gps_ids}

        if len(gps_id_to_cam_id) != len(gps_ids):
            logger.warning('unable to map some GPS to a camera.')

    gps_records = ((timestamp, gps_id, gps_id_to_cam_id[gps_id], gnss_record)
                   for timestamp, gps_id, gnss_record in kapture.flatten(kapture_data.records_gnss)
                   if gps_id in gps_id_to_cam_id)

    for timestamp, gps_id, cam_id, gps_record in gps_records:
        if (timestamp, cam_id) not in kapture_data.records_camera:
            logger.warning(f'no image found corresponding to GPS record ({timestamp}, {cam_id})')
        else:
            image_name = kapture_data.records_camera[timestamp, cam_id]
            image_filepath = get_image_fullpath(kapture_dir_path=kapture_dirpath, image_filename=image_name)
            exif_data = gps_record_to_exif_dict(gps_record)
            update_exif(image_filepath, exif_data)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

import setuptools
import warnings
import os.path as path
from glob import glob
from subprocess import check_call
import os
import tempfile
import sys

HERE = path.normpath(path.dirname(__file__))


# Documentation ########################################################################################################
# setuptools only support md documentation: convert from asciidoc
def read_file(filepath):
    with open(filepath, 'r', encoding='utf-8') as f:
        return f.read()


def read_doc(filepath):
    try:  # from asciidoc to markdown
        # https://tinyapps.org/blog/201701240700_convert_asciidoc_to_markdown.html
        with tempfile.TemporaryDirectory() as tmpdirname:
            xml_filepath = path.join(tmpdirname, 'README.xml')
            md_filepath = path.join(tmpdirname, 'README.md')
            use_shell = sys.platform.startswith("win")
            check_call(['asciidoctor', '-b', 'docbook', filepath, '-o', xml_filepath], shell=use_shell)
            check_call(['pandoc', '-f', 'docbook', '-t', 'markdown_strict', xml_filepath, '-o', md_filepath],
                       shell=use_shell)
            content = read_file(md_filepath)

    except FileNotFoundError:
        warnings.warn('cannot convert asciidoc to markdown.')
        content = read_file(filepath)

    return content


readme_filepath = path.join(HERE, 'README.adoc')
long_description = read_doc(readme_filepath)

########################################################################################################################
setuptools.setup(
    # description
    name='kapture',
    version="1.1.1",
    author="naverlabs",
    author_email="kapture@naverlabs.com",
    description="kapture: file format for SfM",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/naver/kapture/",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],

    # dependencies
    python_requires='>=3.6',
    install_requires=[
        'dataclasses>=0.3;python_version<\'3.7\'',
        'numpy>=1.16',
        'numpy-quaternion>=2019.3.18.14.33.20',
        'numba>=0.42',
        'matplotlib>=3.0',
        'scipy>=1.4',
        'tqdm>=4.30',
        'Pillow>=8.1.1',
        'piexif==1.1.3',
        'requests>=2.21',
        'pyyaml>=5.1',
        'wcmatch>=5.0',
        'tabulate>=0.8.7',
    ],
    extras_require={
        'dev': ['pytest'],
    },
    # sources
    packages=setuptools.find_packages(),
    scripts=[f for f in glob(os.path.join(HERE, 'tools', '*.py'))],
)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Paths manipulations and filesystem operations.
"""

import os
import os.path as path
import shutil
import tempfile
from typing import AnyStr, Iterable, Optional, Union, List

from kapture.utils.logging import getLogger

logger = getLogger()


def path_secure(any_path: AnyStr) -> AnyStr:
    """
    Make sure path representation is OS independent (same on linux and windows),
    because path can be used as an identifier (eg. images).

    :param any_path: path to normalize
    :return: normalized path
    """
    return path.normpath(any_path).replace('\\', '/')


def populate_files_in_dirpath(root_dir_path: str,
                              filename_extensions: Optional[Union[str, List[str]]] = None,
                              do_relative_path: bool = True
                              ) -> Iterable[str]:
    """
    Returns the list of file path into the given root path.
    If a list of extensions is given, returns only files with those extensions.

    :param root_dir_path: the root directory path
    :param filename_extensions: optional file name extensions to filter in
    :param do_relative_path: if True, computes relative path.
    :return: list of paths
    """
    # list all files
    file_paths = (
        path.join(subdir, filename)
        for subdir, _, file_list in os.walk(root_dir_path)
        for filename in file_list
    )
    # if extensions are given, keep only files that complies.
    if filename_extensions:
        if not isinstance(filename_extensions, list):
            # make sure extensions is a list
            filename_extensions = [filename_extensions]
        # make sure given extensions are lower case.
        filename_extensions = [ext.lower() for ext in filename_extensions]
        # check the extension is authorized
        file_paths = (
            file_path for file_path in file_paths
            if path.splitext(file_path)[1].lower() in filename_extensions
        )

    # make it relative to root
    if do_relative_path:
        file_paths = (
            path.relpath(file_path, root_dir_path)
            for file_path in file_paths
        )

    file_paths = (path_secure(file_path)
                  for file_path in file_paths)
    return file_paths


def safe_remove_file(filepath: str, force: bool, comment: Optional[str] = '') -> None:
    """
    Safely remove a file, optionally asking confirmation to the user on the command line.

    :param filepath: path to the file to delete.
    :param force: If True, does not ask the question to the user and just do it.
    :param comment: comment on the file
    """
    if path.isfile(filepath):
        to_delete = (force or
                     (input(f'{comment} {filepath} already exist, would you like to delete it ? [y/N]').lower() == 'y'))
        # delete all or quit
        if to_delete:
            logger.info(f'deleting already existing {filepath}')
            os.remove(filepath)


def safe_remove_any_path(any_path: str, force: bool) -> None:
    """
    Safely remove any file system object (directory, file or link),
    optionally asking confirmation to the user on the command line.

    :param any_path: path to the filesystem object to delete.
    :param force: If True, does not ask the question to the user and just do it.
    """
    if path.exists(any_path):
        to_delete = (force or
                     (input(f'{any_path} already exist, would you like to delete it ? [y/N]').lower() == 'y'))
        # delete all or quit
        if to_delete:
            logger.info(f'deleting already existing {any_path}')
            if path.isfile(any_path) or path.islink(any_path):
                os.remove(any_path)
            else:
                shutil.rmtree(any_path)


def prepend_to_file(file_path: str, text_to_prepend: str) -> None:
    """
    Prepend a string at the beginning of a file.

    :param file_path: path to the file
    :param text_to_prepend: string to prepend
    """
    # First move the file to temp
    temp_copy = tempfile.NamedTemporaryFile(suffix='.txt').name
    shutil.move(file_path, temp_copy)
    try:
        with open(file_path, 'w') as f_dst:
            f_dst.write(text_to_prepend)
            with open(temp_copy, 'r') as f_src:
                while 1:
                    buf = f_src.read(16*1024)
                    if not buf:
                        break
                    f_dst.write(buf)
        os.remove(temp_copy)
    except OSError:
        os.remove(file_path)
        # In case the copy went wrong, restore initial file
        shutil.move(temp_copy, file_path)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Merge kapture objects.
"""

from kapture.io.tar import TarCollection
from typing import List, Optional, Type

import kapture
from kapture.io.records import TransferAction, get_image_fullpath
from kapture.utils.Collections import get_new_if_not_empty

from .merge_reconstruction import merge_keypoints_collections, merge_descriptors_collections
from .merge_reconstruction import merge_global_features_collections, merge_matches_collections
from .merge_reconstruction import merge_points3d_and_observations, merge_points3d
from .merge_records_data import merge_records_data


def merge_table_key1(
        table_list,
        table_constructor,
):
    """
    Merge several table with 1 key (eg. device_id)  into one.
    If multiple entry for a key keep only the first one.

    :param table_list: list of table to merge.
    :param table_constructor: the class type of table.
    :return table_merged
    """
    assert len(table_list) > 0
    table_list = [table for table in table_list if table is not None]
    if not all(isinstance(table, table_constructor) for table in table_list):
        raise TypeError('unexpected type.')
    table_merged = table_constructor()
    for table in table_list:
        for key1, entry in kapture.flatten(table):
            if key1 in table_merged:
                # skip to keep the first one
                continue
            table_merged[key1] = entry
    return table_merged


def merge_table_key2(
        table_list,
        table_constructor,
):
    """
    Merge several table with 2 keys (eg. timestamps, device_id)  into one.
    If multiple entry for a key keep only the first one.

    :param table_list: list of table to merge.
    :param table_constructor: the class type of table.
    :return table_merged
    """
    assert len(table_list) > 0
    table_list = [table for table in table_list if table is not None]
    if not all(isinstance(table, table_constructor) for table in table_list):
        raise TypeError('unexpected type.')
    table_merged = table_constructor()
    for table in table_list:
        for key1, key2, entry in kapture.flatten(table):
            if (key1, key2) in table_merged:
                # skip to keep the first one
                continue
            table_merged[key1, key2] = entry
    return table_merged


def merge_table_key3(
        table_list,
        table_constructor,
        subdict_constructor=dict,
):
    """
    Merge several records lists. Records is a dict (eg. wifi)).
    For record with the same timestamp and sensor identifier, keep only the first one.

    :param table_list: list of table to merge.
    :param table_constructor: the class type of table.
    :param subdict_constructor: used to create a new Dict type
    :return table_merged
    """
    assert len(table_list) > 0
    table_list = [table for table in table_list if table is not None]
    if not all(isinstance(table, table_constructor) for table in table_list):
        raise TypeError('unexpected type.')
    table_merged = table_constructor()
    for table in table_list:
        for key1, key2, key3, entry in kapture.flatten(table):
            if (key1, key2) not in table_merged:
                # if timestamp, sensor_id not there yet, create an instance of dict record
                table_merged[key1, key2] = subdict_constructor()
            table_merged[key1, key2].setdefault(key3, entry)
    return table_merged


def merge_sensors(
        sensors_list: List[Optional[kapture.Sensors]]
) -> kapture.Sensors:
    """
    Merge several sensors lists. For sensor with the same identifier, keep only the first one.

    :param sensors_list: list of sensors
    :return: merge sensors
    """
    return merge_table_key1(
        table_list=sensors_list,
        table_constructor=kapture.Sensors
    )


def merge_rigs(
        rigs_list: List[Optional[kapture.Rigs]]
) -> kapture.Rigs:
    """
    Merge several rigs lists. For sensor with the same rig and sensor identifier, keep only the first one.

    :param rigs_list: list of rigs
    :return: merged rigs
    """
    return merge_table_key2(
        table_list=rigs_list,
        table_constructor=kapture.Rigs
    )


def merge_trajectories(
        trajectories_list: List[Optional[kapture.Trajectories]]
) -> kapture.Trajectories:
    """
    Merge several trajectories lists. For trajectory point with the same timestamp and sensor identifier,
     keep only the first one.

    :param trajectories_list: list of trajectories
    :return: merged trajectories
    """
    return merge_table_key2(
        table_list=trajectories_list,
        table_constructor=kapture.Trajectories
    )


def merge_records_camera(
        records_camera_list: List[Optional[kapture.RecordsCamera]]
) -> kapture.RecordsCamera:
    """
    Merge several camera records lists. For camera record with the same timestamp and sensor identifier,
     keep only the first one.

    :param records_camera_list: list of camera records
    :return: merged camera records
    """
    return merge_table_key2(
        table_list=records_camera_list,
        table_constructor=kapture.RecordsCamera
    )


def merge_records_lidar(
        records_lidar_list: List[Optional[kapture.RecordsLidar]]
) -> kapture.RecordsLidar:
    """
    Merge several lidar records lists. For lidar record with the same timestamp and sensor identifier,
     keep only the first one.

    :param records_lidar_list: list of lidar records
    :return: merged lidar records
    """
    return merge_table_key2(
        table_list=records_lidar_list,
        table_constructor=kapture.RecordsLidar
    )


def merge_records_wifi(
        records_wifi_list: List[Optional[kapture.RecordsWifi]]
) -> kapture.RecordsWifi:
    """
    Merge several wifi records lists.
    For wifi record with the same timestamp, sensor, BSSID,
     keep only the first one.

    :param records_wifi_list: list of wifi records
    :return: merged wifi records
    """
    return merge_table_key3(
        table_list=records_wifi_list,
        table_constructor=kapture.RecordsWifi,
        subdict_constructor=kapture.RecordsWifi.record_type
    )


def merge_records_bluetooth(
        records_bluetooth_list: List[Optional[kapture.RecordsBluetooth]]
) -> kapture.RecordsBluetooth:
    """
    Merge several bluetooth records lists.
    For bluetooth record with the same timestamp, sensor, address,
     keep only the first one.

    :param records_bluetooth_list: list of wifi records
    :return: merged bluetooth records
    """
    return merge_table_key3(
        table_list=records_bluetooth_list,
        table_constructor=kapture.RecordsBluetooth,
        subdict_constructor=kapture.RecordsBluetooth.record_type
    )


def merge_records_gnss(
        records_gnss_list: List[Optional[kapture.RecordsGnss]]
) -> kapture.RecordsGnss:
    """
    Merge several gnss records lists. For gnss record with the same timestamp and sensor identifier,
     keep only the first one.

    :param records_gnss_list: list of gnss records
    :return: merged gnss records
    """
    return merge_table_key2(
        table_list=records_gnss_list,
        table_constructor=kapture.RecordsGnss
    )


def merge_records_accelerometer(
        records_accelerometer_list: List[Optional[kapture.RecordsAccelerometer]]
) -> kapture.RecordsAccelerometer:
    """
    Merge several accelerometer records lists.
    For accelerometer record with the same timestamp and sensor identifier,
     keep only the first one.

    :param records_accelerometer_list: list of accelerometer records
    :return: merged accelerometer records
    """
    return merge_table_key2(
        table_list=records_accelerometer_list,
        table_constructor=kapture.RecordsAccelerometer
    )


def merge_records_gyroscope(
        records_gyroscope_list: List[Optional[kapture.RecordsGyroscope]]
) -> kapture.RecordsGyroscope:
    """
    Merge several gyroscope records lists.
    For gyroscope record with the same timestamp and sensor identifier,
     keep only the first one.

    :param records_gyroscope_list: list of gnss records
    :return: merged gyroscope records
    """
    return merge_table_key2(
        table_list=records_gyroscope_list,
        table_constructor=kapture.RecordsGyroscope
    )


def merge_records_magnetic(
        records_magnetic_list: List[Optional[kapture.RecordsMagnetic]]
) -> kapture.RecordsMagnetic:
    """
    Merge several magnetic records lists.
    For magnetic record with the same timestamp and sensor identifier,
     keep only the first one.

    :param records_magnetic_list: list of gnss records
    :return: merged magnetic records
    """
    return merge_table_key2(
        table_list=records_magnetic_list,
        table_constructor=kapture.RecordsMagnetic
    )


def merge_keep_ids(kapture_list: List[kapture.Kapture],  # noqa: C901: function a bit long but not too complex
                   skip_list: List[Type],
                   data_paths: List[str],
                   tarcollection_list: List[TarCollection],
                   kapture_path: str,
                   images_import_method: TransferAction) -> kapture.Kapture:
    """
    Merge multiple kapture while keeping ids (sensor_id) identical in merged and inputs.

    :param kapture_list: list of kapture to merge.
    :param skip_list: optional types not to merge. sensors and rigs are unskippable
    :param data_paths: list of path to root path directory in same order as mentioned in kapture_list.
    :param tarcollection_list: list of opened tar archives same order as mentioned in kapture_list.
    :param kapture_path: directory root path to the merged kapture.
    :param images_import_method: method to transfer image files
    :return: merged kapture
    """
    merged_kapture = kapture.Kapture()

    # get the union of all sensors
    new_sensors = merge_sensors([every_kapture.sensors for every_kapture in kapture_list])
    # if merge_sensors returned an empty object, keep merged_kapture.sensors to None
    merged_kapture.sensors = get_new_if_not_empty(new_sensors, merged_kapture.sensors)

    # get the union of all rigs
    new_rigs = merge_rigs([every_kapture.rigs for every_kapture in kapture_list])
    # if merge_rigs returned an empty object, keep merged_kapture.rigs to None
    merged_kapture.rigs = get_new_if_not_empty(new_rigs, merged_kapture.rigs)

    # all fields below can be skipped with skip_list
    # we do not assign the properties when the merge evaluate to false, we keep it as None
    if kapture.Trajectories not in skip_list:
        new_trajectories = merge_trajectories([every_kapture.trajectories for every_kapture in kapture_list])
        merged_kapture.trajectories = get_new_if_not_empty(new_trajectories, merged_kapture.trajectories)

    if kapture.RecordsCamera not in skip_list:
        new_records_camera = merge_records_camera([every_kapture.records_camera for every_kapture in kapture_list])
        merged_kapture.records_camera = get_new_if_not_empty(new_records_camera, merged_kapture.records_camera)

        merge_records_data([[image_name
                             for _, _, image_name in kapture.flatten(every_kapture.records_camera)]
                            if every_kapture.records_camera is not None else []
                            for every_kapture in kapture_list],
                           [get_image_fullpath(data_path, image_filename=None) for data_path in data_paths],
                           kapture_path,
                           images_import_method)

    if kapture.RecordsLidar not in skip_list:
        new_records_lidar = merge_records_lidar([every_kapture.records_lidar
                                                 for every_kapture in kapture_list])
        merged_kapture.records_lidar = get_new_if_not_empty(new_records_lidar,
                                                            merged_kapture.records_lidar)
    if kapture.RecordsWifi not in skip_list:
        new_records_wifi = merge_records_wifi([every_kapture.records_wifi
                                               for every_kapture in kapture_list])
        merged_kapture.records_wifi = get_new_if_not_empty(new_records_wifi,
                                                           merged_kapture.records_wifi)
    if kapture.RecordsBluetooth not in skip_list:
        new_records_bluetooth = merge_records_bluetooth([every_kapture.records_bluetooth
                                                         for every_kapture in kapture_list])
        merged_kapture.records_bluetooth = get_new_if_not_empty(new_records_bluetooth,
                                                                merged_kapture.records_bluetooth)
    if kapture.RecordsGnss not in skip_list:
        new_records_gnss = merge_records_gnss([every_kapture.records_gnss
                                               for every_kapture in kapture_list])
        merged_kapture.records_gnss = get_new_if_not_empty(new_records_gnss,
                                                           merged_kapture.records_gnss)
    if kapture.RecordsAccelerometer not in skip_list:
        new_records_accelerometer = merge_records_accelerometer([every_kapture.records_accelerometer
                                                                 for every_kapture in kapture_list])
        merged_kapture.records_accelerometer = get_new_if_not_empty(new_records_accelerometer,
                                                                    merged_kapture.records_accelerometer)
    if kapture.RecordsGyroscope not in skip_list:
        new_records_gyroscope = merge_records_gyroscope([every_kapture.records_gyroscope
                                                         for every_kapture in kapture_list])
        merged_kapture.records_gyroscope = get_new_if_not_empty(new_records_gyroscope,
                                                                merged_kapture.records_gyroscope)
    if kapture.RecordsMagnetic not in skip_list:
        new_records_magnetic = merge_records_magnetic([every_kapture.records_magnetic
                                                       for every_kapture in kapture_list])
        merged_kapture.records_magnetic = get_new_if_not_empty(new_records_magnetic,
                                                               merged_kapture.records_magnetic)

    # for the reconstruction, except points and observations, the files are copied with shutil.copy
    # if kapture_path evaluates to False, all copies will be skipped (but classes will be filled normally)
    if kapture.Keypoints not in skip_list:
        keypoints = [every_kapture.keypoints for every_kapture in kapture_list]
        keypoints_not_none = [k for k in keypoints if k is not None]
        if len(keypoints_not_none) > 0:
            new_keypoints = merge_keypoints_collections(keypoints, data_paths, kapture_path, tarcollection_list)
            merged_kapture.keypoints = get_new_if_not_empty(new_keypoints, merged_kapture.keypoints)
    if kapture.Descriptors not in skip_list:
        descriptors = [every_kapture.descriptors for every_kapture in kapture_list]
        descriptors_not_none = [k for k in descriptors if k is not None]
        if len(descriptors_not_none) > 0:
            new_descriptors = merge_descriptors_collections(descriptors, data_paths, kapture_path, tarcollection_list)
            merged_kapture.descriptors = get_new_if_not_empty(new_descriptors, merged_kapture.descriptors)
    if kapture.GlobalFeatures not in skip_list:
        global_features = [every_kapture.global_features for every_kapture in kapture_list]
        global_features_not_none = [k for k in global_features if k is not None]
        if len(global_features_not_none) > 0:
            new_global_features = merge_global_features_collections(global_features, data_paths,
                                                                    kapture_path, tarcollection_list)
            merged_kapture.global_features = get_new_if_not_empty(new_global_features, merged_kapture.global_features)
    if kapture.Matches not in skip_list:
        matches = [every_kapture.matches for every_kapture in kapture_list]
        matches_not_none = [k for k in matches if k is not None]
        if len(matches_not_none) > 0:
            new_matches = merge_matches_collections(matches, data_paths, kapture_path, tarcollection_list)
            merged_kapture.matches = get_new_if_not_empty(new_matches, merged_kapture.matches)
    if kapture.Points3d not in skip_list and kapture.Observations not in skip_list:
        points_and_obs = [(every_kapture.points3d, every_kapture.observations) for every_kapture in kapture_list]
        new_points, new_observations = merge_points3d_and_observations(points_and_obs)
        merged_kapture.points3d = get_new_if_not_empty(new_points, merged_kapture.points3d)
        merged_kapture.observations = get_new_if_not_empty(new_observations, merged_kapture.observations)
    elif kapture.Points3d not in skip_list:
        points = [every_kapture.points3d for every_kapture in kapture_list]
        new_points = merge_points3d(points)
        merged_kapture.points3d = get_new_if_not_empty(new_points, merged_kapture.points3d)
    return merged_kapture



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This script imports images from a folder in the kapture format
"""

import argparse
import logging
import os
import os.path as path
import PIL
from PIL import Image

# kapture
import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.logging
from kapture.io.structure import delete_existing_kapture_files
from kapture.io.csv import kapture_to_dir
from kapture.utils.paths import path_secure
from kapture.io.records import TransferAction, import_record_data_from_dir_auto

logger = logging.getLogger('image_folder')


def import_image_folder(images_path: str,
                        kapture_path: str,
                        force_overwrite_existing: bool = False,
                        images_import_method: TransferAction = TransferAction.skip) -> None:
    """
    Imports the images of a folder to a kapture. This creates only images and cameras.

    :param images_path: path to directory containing the images.
    :param kapture_path: path to kapture root directory.
    :param force_overwrite_existing: Silently overwrite kapture files if already exists.
    :param images_import_method: choose how to import actual image files.
    """
    os.makedirs(kapture_path, exist_ok=True)
    delete_existing_kapture_files(kapture_path, force_erase=force_overwrite_existing)

    cameras = kapture.Sensors()
    images = kapture.RecordsCamera()

    file_list = [os.path.relpath(os.path.join(dirpath, filename), images_path)
                 for dirpath, dirs, filenames in os.walk(images_path)
                 for filename in filenames]
    file_list = sorted(file_list)

    logger.info('starting conversion...')
    for n, filename in enumerate(file_list):
        # test if file is a valid image
        try:
            # lazy load
            with Image.open(path.join(images_path, filename)) as im:
                width, height = im.size
                model_params = [width, height]
        except (OSError, PIL.UnidentifiedImageError):
            # It is not a valid image: skip it
            logger.info(f'Skipping invalid image file {filename}')
            continue

        camera_id = f'sensor{n}'
        images[(n, camera_id)] = path_secure(filename)  # don't forget windows
        cameras[camera_id] = kapture.Camera(kapture.CameraType.UNKNOWN_CAMERA, model_params)

    # import (copy) image files.
    logger.info('import image files ...')
    filename_list = [f for _, _, f in kapture.flatten(images)]
    import_record_data_from_dir_auto(images_path, kapture_path, filename_list, images_import_method)

    # pack into kapture format
    imported_kapture = kapture.Kapture(sensors=cameras, records_camera=images)
    logger.info('writing imported data...')
    kapture_to_dir(kapture_path, imported_kapture)


def import_image_folder_command_line() -> None:
    """
    Do the image list import to kapture using the parameters given on the command line.
    """
    parser = argparse.ArgumentParser(description='imports images from a folder in the kapture format')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete output if already exists.')
    # import ###########################################################################################################
    parser.add_argument('-i', '--input', required=True, help='input path to images root folder')
    parser.add_argument('-o', '--output', required=True, help='output directory')
    parser.add_argument('--image_transfer', type=TransferAction, default=TransferAction.link_absolute,
                        help=f'How to import images [link_absolute], '
                             f'choose among: {", ".join(a.name for a in TransferAction)}')
    ####################################################################################################################
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    import_image_folder(args.input, args.output, args.force, args.image_transfer)


if __name__ == '__main__':
    import_image_folder_command_line()



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
script to easily  download kapture datasets.
"""

import argparse
import logging
import sys
import os
import os.path as path
import requests
import yaml
import fnmatch
from typing import Dict, Optional, List
from tqdm import tqdm
import path_to_kapture  # noqa: F401
# import kapture
import kapture.utils.logging
from kapture.converter.downloader.download import download_file, get_remote_file_size
from kapture.converter.downloader.archives import untar_file, compute_sha256sum
from kapture.io.csv import get_version_from_csv_file
from kapture.utils.upgrade import upgrade_1_0_to_1_1_inplace

logger = logging.getLogger('downloader')
logging.basicConfig(format='%(levelname)-8s::%(name)s: %(message)s')

INDEX_FILENAME = 'kapture_dataset_index.yaml'
DEFAULT_DATASET_PATH = path.normpath(path.abspath('.'))
DEFAULT_REPOSITORY_URL = 'https://github.com/naver/kapture/raw/main/dataset'
# DEFAULT_REPOSITORY_URL = 'https://download.europe.naverlabs.com/kapture/'


def ask_confirmation(question):
    """
    ask "question" to the user.
    The "answer" return value is True for "yes" or False for "no".
    """
    validate = ['yes', 'y', 'ye']
    prompt = ' [y/N]\n'
    sys.stdout.write(question + prompt)
    user_choice = input().lower()
    return user_choice in validate


class Dataset:
    def __init__(
            self,
            name: str,
            install_dirpath: str,
            archive_url: str,
            archive_sha256sum: str,
            install_script_filename: Optional[str] = None
    ):
        """
        :param name: name of the archive (dataset or part of a dataset)
        :param install_dirpath: input absolute path to root directory where all datasets are installed.
        :param archive_url: remote url of the dataset archive (tar).
        :param archive_sha256sum: expected sha256 sum of the archive file.
        :param install_script_filename: if given, this script is to be called to finish installation (eg. dl 3rd party).
        """
        self._name = name

        self._install_local_path = install_dirpath
        self._archive_filepath = path.join(install_dirpath, name + '.tar.gz')
        self._dataset_index_filepath = path.join(install_dirpath, 'kapture_dataset_index.yaml')
        self._dataset_install_list_filepath = path.join(install_dirpath, 'kapture_dataset_installed.yaml')
        self._archive_url = archive_url
        self._archive_sha256sum_remote = archive_sha256sum
        self._install_script_filename = install_script_filename
        self._is_installed = None

    def mark_as_installed(self, installed=True):
        # load previous version
        if path.isfile(self._dataset_install_list_filepath):
            with open(self._dataset_install_list_filepath, 'rt') as f:
                datasets_list = set(yaml.safe_load(f))
        else:
            datasets_list = set()

        if installed:
            datasets_list.add(self._name)
        else:
            datasets_list.discard(self._name)

        # update with current dataset status
        # write updated version
        logger.debug(f'saving {"installed" if installed else "not installed"} in {self._dataset_install_list_filepath}')
        with open(self._dataset_install_list_filepath, 'wt') as f:
            yaml.dump(list(datasets_list), f)

        self._is_installed = installed

    def is_installed(self, installation_list_cache=None) -> bool:
        """
        :param installation_list_cache: spare the read of the yaml file if already loaded before.
        """
        if self._is_installed is not None:
            return self._is_installed

        if installation_list_cache is not None:
            installation_list = installation_list_cache
        elif path.isfile(self._dataset_install_list_filepath):
            with open(self._dataset_install_list_filepath, 'rt') as f:
                installation_list = yaml.safe_load(f)
        else:
            installation_list = []

        assert isinstance(installation_list, list)
        self._is_installed = self._name in installation_list
        return self._is_installed

    def is_sha256_consistent(self) -> bool:
        """ check sha256sum of the archive against the expected sha256. Returns true if they are the same. """
        if not path.isfile(self._archive_filepath):
            return False
        # size is consistent, check sha256
        logger.debug(f'checking sha256sum of {path.basename(self._archive_filepath)}, may take a while...')
        sha256sum_archive_local = compute_sha256sum(self._archive_filepath)
        if sha256sum_archive_local != self._archive_sha256sum_remote:
            logger.warning(f'sha256sum discrepancy for {self._archive_filepath} :\n'
                           f'\tlocal :{sha256sum_archive_local}\n'
                           f'\tremote:{self._archive_sha256sum_remote}')
            return False
        logger.debug(f'{path.basename(self._archive_filepath)} is valid (sha256sum)')
        return True

    @property
    def url(self) -> str:
        return self._archive_url

    def prob_status(self, check_online=False) -> str:
        """
        gives the actual dataset status
         - not installed: means is not installed (wo info about the server)
         - online: means not installed, not downloaded, but reachable.
         - not reachable: means not installed, not downloaded, but NOT reachable.
         - installed: means has been downloaded and installed (extracted).
         - downloaded: means has been downloaded (tar) but not installed (extracted) yet.
         - incomplete: means partially downloaded
         - corrupted: means that the downloaded archive is corrupted (inconsistent size or sh256).
         - unknown: should not happen

        :param check_online: If true, will ping the url to check its actually online
        """

        probing_status = None
        size_archive_local = size_archive_online = -1
        if self.is_installed():
            # yaml file says its installed, trust it (no other choices).
            probing_status = 'installed'

        if probing_status is None and not path.isfile(self._archive_filepath):
            # not installed, no archive there, check its online
            if not check_online:
                probing_status = 'not installed'
            else:
                remote_file_size_in_bytes = get_remote_file_size(url=self._archive_url)
                if remote_file_size_in_bytes is not None:
                    probing_status = f'online {remote_file_size_in_bytes/(1024*1024):5.0f} MB '
                else:
                    probing_status = 'not reachable'

        # not installed, but archive there, check 1) its incomplete or 2) corrupted. If neither, is just downloaded.
        if probing_status is None:
            assert path.isfile(self._archive_filepath)
            # 1) check its incomplete: has it proper size ?
            size_archive_online = get_remote_file_size(self._archive_url)
            size_archive_local = int(path.getsize(self._archive_filepath))
            if size_archive_online is None:
                logger.critical('impossible to retrieve remote file size.')
                probing_status = 'corrupted'
            elif size_archive_local > size_archive_online:
                logger.critical(f'inconsistent file size, file is bigger than it is supposed to be'
                                f' ({size_archive_online} vs {size_archive_local}).')
                probing_status = 'corrupted'
            elif size_archive_local < size_archive_online:
                logger.debug(f'file_size_online={size_archive_online} != file_size_local={size_archive_local}')
                probing_status = 'incomplete'

        if probing_status is None:
            # 2) size is consistent, check sha256
            assert size_archive_local == size_archive_online
            if not self.is_sha256_consistent():
                probing_status = 'corrupted'

        # archive is there, not installed, with valid sha256 then it must be just downloaded.
        if probing_status is None:
            probing_status = 'downloaded'

        return probing_status

    def __repr__(self) -> str:
        return f'{self.prob_status():10} | {self._name:30} | {self._archive_url}'

    def download(self,
                 force_overwrite: bool = False,
                 nb_attempt: int = 2,
                 previous_status=None) -> str:
        status = previous_status or self.prob_status()
        if force_overwrite and path.isfile(self._archive_filepath):
            logger.debug('remove previously downloaded file and start from scratch.')
            os.remove(self._archive_filepath)
            status = self.prob_status()

        if status == 'downloaded':
            logger.info(f'{path.basename(self._archive_filepath)} is already downloaded.')
            return status

        for attempt in range(nb_attempt):
            if status == 'downloaded':
                logger.debug(f'{path.basename(self._archive_filepath)} is downloaded')
                break
            # check archive file integrity
            if status == 'corrupted':
                logger.warning(f'archive {path.basename(self._archive_filepath)} is corrupted. '
                               f'It will be downloaded again.')
                # if corrupted: remove the archive and start over
                os.remove(self._archive_filepath)
            logger.debug(f'downloading {path.basename(self._archive_filepath)} (attempt {attempt + 1}/{nb_attempt})')
            download_file(self._archive_url, self._archive_filepath)
            status = self.prob_status()
        return status

    def install(self,
                force_overwrite: bool = False,
                no_cleaning: bool = False) -> str:
        """
        Install handles download and untar, and possibly tar remove and launch of an install script.

        """
        # make sure self._status is up to date.
        if force_overwrite:
            self.mark_as_installed(False)

        status = self.prob_status()
        if status == 'installed':
            logger.info(f'{self._install_local_path} already exists: skipped')
            return status

        # download (lets try it twice
        status = self.download(previous_status=status)

        if status != 'downloaded':
            # if still not downloaded, it means it failed to download
            logger.critical(f'failed to download {self._name}')
            return status

        # deflating
        logger.debug(f'deflating {path.basename(self._archive_filepath)} to {self._install_local_path}')
        untar_file(self._archive_filepath, self._install_local_path)

        # optionally: script ?
        if self._install_script_filename is not None:
            logger.debug(f'installation script  {self._install_script_filename}')
            install_script_filepath = path.join(self._install_local_path, self._install_script_filename)
            logger.debug(f'applying installation script {install_script_filepath}')
            os.system(install_script_filepath)
        # optionally: clean tar.gz file
        if not no_cleaning:
            logger.debug(f'removing archive file {path.basename(self._archive_filepath)}')
            os.remove(self._archive_filepath)

        # done
        self.mark_as_installed()
        self.upgrade()
        logger.debug(f'done installing {self._name}')
        status = self.prob_status()
        return status

    def upgrade(self):
        # upgrade ?
        status = self.prob_status()
        if status == 'installed':
            # list all sensors.txt files
            file_list = [os.path.join(dp, f) for dp, dn, filenames in os.walk(self._install_local_path)
                         for f in filenames if f == 'sensors.txt']
            upgrade_successful = False
            # check version
            for filename in file_list:
                version = get_version_from_csv_file(filename)
                if version is None or version == '1.0':
                    logger.debug(f'upgrading {filename} from version {version} to 1.1')
                    kapture_path = path.dirname(path.dirname(filename))
                    # run inplace upgrade script
                    upgrade_1_0_to_1_1_inplace(kapture_path, None, None, None, 'L2', 'L2')
                    logger.info(f'{kapture_path} - successfully upgraded to 1.1')
                elif version == '1.1':
                    logger.info(f'{filename} already in 1.1, no need to upgrade')
                else:
                    logger.warning(f'{filename} - version {version} is unknown')
            return upgrade_successful
        else:
            logger.warning('dataset not yet installed, cannot attempt upgrade')
            return False


def load_datasets_from_index(
        index_filepath: str,
        install_path: str,
        filter_patterns: Optional[List[str]] = None
) -> Dict[str, Dataset]:
    """
    Parses and load data from the index files, under yaml format.
    the yaml file looks like :
    ----
    robotcar_seasons_02:
      url: http://download.europe.naverlabs.com//kapture/robotcar_seasons_02.tar.gz
      sha256sum: 542ef47c00d5e387cfb0dcadb2459ae2fb17d59010cc51bae0c49403b4fa6a18
    ----

    :param index_filepath: input absolute path to index file
    :param install_path: input absolute path to install directory
    :param filter_patterns: optional input list of unix-like patterns (e.g. SiLDa*) to filter datasets
    :return: dict name -> [url, sub_path, sha256sum]
    """
    if not path.isfile(index_filepath):
        raise FileNotFoundError('no index file: do an update.')
    with open(index_filepath, 'rt') as f:
        datasets_yaml = yaml.safe_load(f)

    if len(datasets_yaml) == 0:
        raise FileNotFoundError('invalid index file: do an update.')

    nb_total = len(datasets_yaml)
    # filter only dataset matching filter_patterns
    if filter_patterns:
        datasets_yaml = {dataset_name: data_yaml
                         for dataset_name, data_yaml in datasets_yaml.items()
                         if any(fnmatch.fnmatch(dataset_name, pattern) for pattern in filter_patterns)}

    logger.debug(f'will prob status for {len(datasets_yaml)}/{nb_total} datasets ...')
    datasets = {}

    hide_progress_bar = True or logger.getEffectiveLevel() > logging.INFO
    for dataset_name, data_yaml in tqdm(datasets_yaml.items(), disable=hide_progress_bar):
        datasets[dataset_name] = Dataset(name=dataset_name,
                                         install_dirpath=install_path,
                                         archive_url=data_yaml['url'],
                                         archive_sha256sum=data_yaml['sha256sum'],
                                         install_script_filename=data_yaml.get('install_script'))

    return datasets


def kapture_download_dataset(args, index_filepath: str):
    """
    Do the dataset download command
    """
    global_status = 0
    if args.cmd == 'update':
        logger.info(f'updating dataset list from {args.repo} ...')
        index_remote_url = args.repo + '/' + INDEX_FILENAME
        logger.debug(f'retrieving index at {index_remote_url}')
        r = requests.get(index_remote_url, allow_redirects=True)
        if r.status_code != requests.codes.ok:
            raise ConnectionError(f'unable to grab {index_remote_url} (code:{r.status_code})')
        with open(index_filepath, 'wt') as f:
            f.write(r.text)
        datasets = load_datasets_from_index(index_filepath=index_filepath,
                                            install_path=args.install_path)
        logger.info(f'dataset index retrieved successfully: {len(datasets)} datasets')

    elif args.cmd == 'list':
        logger.info(f'listing dataset {index_filepath} ...')
        datasets = load_datasets_from_index(index_filepath=index_filepath,
                                            install_path=args.install_path,
                                            filter_patterns=args.dataset)
        for name, dataset in datasets.items():
            status = dataset.prob_status(check_online=args.full)
            if status == "not reachable" or status == "incomplete" or status == "corrupted":
                global_status = 1
            print(f'{status:^16}| {name:40} | {dataset.url}')

    elif args.cmd == 'install':
        logger.debug(f'will install dataset: {args.dataset} ...')
        dataset_index = load_datasets_from_index(index_filepath=index_filepath,
                                                 install_path=args.install_path,
                                                 filter_patterns=args.dataset)
        if len(dataset_index) == 0:
            raise ValueError('There is no matching dataset.'
                             ' Make sure you used quotes (") to prevent shell interpreting * wildcard.')

        logger.info(f'{len(dataset_index)} dataset will be installed.')
        for name, dataset in dataset_index.items():
            logger.info(f'{name}: starting installation  ...')
            status = dataset.install(force_overwrite=args.force, no_cleaning=args.no_cleaning)
            logger.info(f'{name} install: ' + 'successful' if status == 'installed' else 'failed')

    elif args.cmd == 'upgrade':
        logger.debug(f'will install dataset: {args.dataset} ...')
        dataset_index = load_datasets_from_index(index_filepath=index_filepath,
                                                 install_path=args.install_path,
                                                 filter_patterns=args.dataset)
        if len(dataset_index) == 0:
            raise ValueError('There is no matching dataset.'
                             ' Make sure you used quotes (") to prevent shell interpreting * wildcard.')

        logger.info(f'{len(dataset_index)} dataset will be installed.')
        for name, dataset in dataset_index.items():
            logger.info(f'{name}: starting installation  ...')
            dataset.upgrade()

    elif args.cmd == 'download':
        logger.debug(f'will download dataset: {args.dataset} ...')
        dataset_index = load_datasets_from_index(index_filepath=index_filepath,
                                                 install_path=args.install_path,
                                                 filter_patterns=args.dataset)
        if len(dataset_index) == 0:
            raise ValueError('There is no matching dataset.'
                             ' Make sure you used quotes (") to prevent shell interpreting * wildcard.')
        logger.info(f'{len(dataset_index)} dataset will be downloaded.')
        for name, dataset in dataset_index.items():
            logger.info(f'downloading {name} ...')
            dataset.download(force_overwrite=args.force)

    return global_status


def kapture_download_dataset_cli():
    """
    Parse the kapture_download_dataset command line .
    """
    parser = argparse.ArgumentParser(description='download kapture datasets.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.INFO, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('--install_path', default=path.normpath(path.join(DEFAULT_DATASET_PATH)),
                        help=f'path to index files listing all datasets'
                             f' [{path.normpath(path.join(DEFAULT_DATASET_PATH))}]')
    subparsers = parser.add_subparsers(help='sub-command help', dest='cmd')
    ####################################################################################################################
    # create the parser for the "update" command
    parser_update = subparsers.add_parser('update', help='update dataset index')
    parser_update.set_defaults(cmd='update')
    parser_update.add_argument('--repo', default=DEFAULT_REPOSITORY_URL,
                               help='url of the repository.')
    ####################################################################################################################
    parser_list = subparsers.add_parser('list', help='display dataset index')
    parser_list.set_defaults(cmd='list')
    parser_list.add_argument('dataset', nargs='*', default=[])
    parser_list.add_argument('--full', action='store_true', default=False,
                             help='Display status and size of remote dataset.')
    ####################################################################################################################
    parser_install = subparsers.add_parser('install', help='install dataset')
    parser_install.set_defaults(cmd='install')
    parser_install.add_argument('-f', '--force', action='store_true', default=False,
                                help='Force installation even if dataset has already been installed.')
    parser_install.add_argument('--no_cleaning', action='store_true', default=False,
                                help='Do not delete downloaded tar.gz file.')
    parser_install.add_argument('dataset', nargs='*', default=[],
                                help='name of the dataset to download. Can use unix-like wildcard.')
    ####################################################################################################################
    parser_download = subparsers.add_parser('download', help='download dataset, without installing it')
    parser_download.set_defaults(cmd='download')
    parser_download.add_argument('-f', '--force', action='store_true', default=False,
                                 help='Force installation even if dataset has already been installed.')
    parser_download.add_argument('dataset', nargs='*', default=[],
                                 help='name of the dataset to download. Can use unix-like wildcard.')
    ####################################################################################################################
    parser_install = subparsers.add_parser('upgrade', help='upgrade dataset to 1.1')
    parser_install.set_defaults(cmd='upgrade')
    parser_install.add_argument('dataset', nargs='*', default=[],
                                help='name of the dataset to download. Can use unix-like wildcard.')
    ####################################################################################################################

    args = parser.parse_args()

    logger.setLevel(args.verbose or logging.INFO)
    logger.debug(f'{sys.argv[0]} \\\n' + '  \\\n'.join(
        '--{:20} {:100}'.format(k, str(v))
        for k, v in vars(args).items()))

    try:
        index_filepath = path.join(args.install_path, INDEX_FILENAME)
        if not args.cmd:
            parser.print_help()
            logger.critical(f'Choose command among [ {" | ".join(subparsers.choices)} ]')
            exit(-1)
        return kapture_download_dataset(args, index_filepath)

    except Exception as e:
        logger.critical(e)
        # raise e
        return -1


if __name__ == '__main__':
    sys.exit(kapture_download_dataset_cli())



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Script to import a rosbag with images from several camera and position associated to the images recorded.
Works by default with the RealSense T265 camera.
"""

import argparse
import logging
from os import path
import sys
# kapture
import path_to_kapture  # enables import kapture  # noqa: F401
import kapture.io.csv as kcsv
import kapture.utils.logging

# rosbag
try:
    import rosbag  # noqa: F401
    from kapture.converter.ros_tools.import_rosbag import RosBagImporter

    has_rosbag = True
except ModuleNotFoundError:
    has_rosbag = False

logger = logging.getLogger('rosbag')  # Using global logger

# Default ROS topics: RealSense T265 camera
TOPIC_ODOMETRY = '/camera/odom/sample'
TOPIC_CAMERA_LEFT = '/camera/fisheye1/image_raw'
TOPIC_CAMERA_RIGHT = '/camera/fisheye2/image_raw'
DEFAULT_IMAGE_TOPICS = [TOPIC_CAMERA_LEFT, TOPIC_CAMERA_RIGHT]


def import_rosbag(args) -> None:
    """
    Import the rosbag with the parameters.

    :param args: arguments to use
    """
    # Import predefined rig and sensors list
    rig_sensors_kapture_path = path.abspath(args.kapture_rig)
    kapture_data = kcsv.kapture_from_dir(rig_sensors_kapture_path)
    rigs = kapture_data.rigs
    rig_id = None
    if args.odometry_topic is not None:
        # If we have trajectory points, we need a rig with a unique id
        if kapture_data.rigs is None or len(kapture_data.rigs) == 0:
            raise ValueError(f'Rig definition is empty in {rig_sensors_kapture_path}')
        for r_id in rigs.keys():
            if rig_id is None:
                rig_id = r_id
            elif r_id != rig_id:
                raise ValueError(f'Found rig_id {r_id} and {rig_id}:'
                                 f' there should be only one rig_id defined in {rig_sensors_kapture_path}')
        if rig_id is None:
            raise ValueError(f'No rig defined in {kapture_data}')
    if kapture_data.cameras is None or len(kapture_data.cameras) == 0:
        raise ValueError(f'Cameras definition is empty in {rig_sensors_kapture_path}')

    importer = RosBagImporter(args.bag_file, rigs, kapture_data.sensors, args.kapture_output, args.force)
    # Import the images
    image_topics = args.image_topics
    if not image_topics:
        image_topics = DEFAULT_IMAGE_TOPICS
    importer.import_multi_camera(args.odometry_topic, image_topics, args.camera_identifiers,
                                 args.all_pose, args.match_pose, args.percent)
    importer.save_to_kapture(rig_id)


def import_rosbag_command_line() -> None:
    """
    Do the rosbag to kapture import using the command line parameters provided by the user.
    """
    parser = argparse.ArgumentParser(description='Imports ROS bag file to kapture format.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument('-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
                                  action=kapture.utils.logging.VerbosityParser,
                                  help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument('-q', '--silent', '--quiet', action='store_const', dest='verbose',
                                  const=logging.CRITICAL)
    parser.add_argument('-y', '--force', action='store_true', default=False,
                        help='silently delete kapture data if already exists.')
    # create the parser for the import command #########################################################################
    parser.add_argument('-b', '--bag_file', default='input.bag',
                        help='bag file path', required=False)
    parser.add_argument('-r', '--kapture_rig',
                        help='kapture with rig and sensors definitions for the capturing equipment '
                             'corresponding to the ROS bag', required=True)
    parser.add_argument('-o', '--odometry_topic', default=TOPIC_ODOMETRY,
                        help=f'odometry topic to import, defaults to {TOPIC_ODOMETRY}', required=False)
    parser.add_argument('-i', '--image_topics', nargs="*",
                        help=f'list of image topics to import, defaults to {DEFAULT_IMAGE_TOPICS}',
                        required=False)
    parser.add_argument('-c', '--camera_identifiers', nargs="*",
                        help='list of camera identifiers as defined in the sensors file of the kapture rig'
                             ' and corresponding to the image topics in the same order;',
                        required=True)
    parser.add_argument('-a', '--all_pose', action='store_true', default=False,
                        help='Save all poses from the odometry topic to the trajectory')
    parser.add_argument('-m', '--match_pose', action='store_true', default=True,
                        help='Find a matching pose for every image, and saved it in the trajectory'
                             ' with the same timestamp')
    parser.add_argument('-p', '--percent', type=int, default=100,
                        help='percentage of images to keep', required=False)
    parser.add_argument('-k', '--kapture_output', required=True,
                        help='directory where to save Kapture files.')
    ####################################################################################################################
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # for debug, let kapture express itself.
        kapture.utils.logging.getLogger().setLevel(args.verbose)
    logger.debug(f'{sys.argv[0]} \\\n' + '  \\\n'.join(
        '--{:20} {:100}'.format(k, str(v))
        for k, v in vars(args).items()))

    import_rosbag(args)


if __name__ == '__main__':
    if not has_rosbag:
        logger.fatal("You don't have ROS or import_rosbag installed")
        sys.exit(-1)
    try:
        import_rosbag_command_line()
    except (AssertionError, OSError, ValueError) as any_ex:
        logger.fatal(f'Fatal error: {any_ex}')



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Virtual gallery constants.
"""

import kapture

virtual_gallery_camera_model = kapture.CameraType.PINHOLE
virtual_gallery_width = 1920
virtual_gallery_height = 1080



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Additional utility functions and classes used by kapture but not specific to it.
"""

from .Collections import AutoEnum  # noqa: F401
from .paths import path_secure, populate_files_in_dirpath  # noqa: F401
